/*
 * Copyright (c) 2014-2020 by Farsight Security, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef ASINFO_H_INCLUDED
#define ASINFO_H_INCLUDED 1

#include <stdbool.h>

#ifndef CRIPPLED_LIBC
char *
asinfo_from_rr(const char *rrtype, const char *rdata, char **asn, char **cidr);
#endif

bool
asinfo_domain_exists(const char *);

void
asinfo_shutdown(void);

#endif /*ASINFO_H_INCLUDED*/
